#include "directory_test.h"

START

TEST(Directory, Exists) {
  EXPECT_FALSE(directory_exists("/catdog"));
  EXPECT_TRUE(directory_exists("/"));
  EXPECT_TRUE(directory_exists("/etc"));
}

TEST(Directory, IsEmpty) {
  char *_dirname2 = null;
  TempFilename *_dirname1 = null;

  EXPECT_FALSE(directory_is_empty("/"));
  EXPECT_FALSE(directory_is_empty("/etc"));

  temp_filename_new(&_dirname1, "cat");
  string_new_concat(&_dirname2, _dirname1->_filename, "/dog");

  directory_ensure_exists(_dirname1->_filename);
  temp_filename_unlock(_dirname1);
  EXPECT_TRUE(directory_is_empty(_dirname1->_filename));
  directory_remove(_dirname1->_filename);

  directory_ensure_exists(_dirname2);
  EXPECT_TRUE(directory_is_empty(_dirname2));
  directory_remove(_dirname2);

  string_delete(&_dirname2);
  temp_filename_delete(&_dirname1);
}

TEST(Directory, Read) {
  char *_dirname = null;
  char *_filename = null;
  char **_entries = null;
  ulong n;
  TempFilename *_s = null;

  temp_filename_new(&_s, null);
  string_new_substring(&_dirname, _s->_filename, 0,
      string_nth_index_of(_s->_filename, 2, '/') + 1);
  string_renew_append(&_dirname,
      _s->_filename + string_nth_index_of(_s->_filename, 4, '/') + 1);
  temp_filename_delete(&_s);
  directory_ensure_exists(_dirname);

  n = directory_read(
      _dirname,
      null,
      null,
      false,
      false,
      &_entries);
  EXPECT_ULEQ(0, n);
  delete(&_entries, n, string_delete);

  string_new_concat(&_filename, _dirname, "/file1.txt");
  file_create(_filename);
  n = directory_read(
      _dirname,
      null,
      null,
      false,
      false,
      &_entries);
  EXPECT_ULEQ(1, n);
  EXPECT_STREQ("file1.txt", _entries[0]);
  delete(&_entries, n, string_delete);

  n = directory_read(
      _dirname,
      null,
      null,
      false,
      true,
      &_entries);
  EXPECT_ULEQ(1, n);
  EXPECT_STREQ(_filename, _entries[0]);
  delete(&_entries, n, string_delete);

  file_remove(_filename);
  string_delete(&_filename);

  directory_remove(_dirname);
  string_delete(&_dirname);
}

TEST(Directory, TryRead) {
  char *_filename = null;
  char **_entries = null;
  ulong n;
  TempFilename *_dirname = null;

  temp_filename_new(&_dirname, null);
  directory_ensure_exists(_dirname->_filename);
  temp_filename_unlock(_dirname);

  EXPECT_TRUE(
      directory_try_read(
        _dirname->_filename,
        null,
        null,
        false,
        false,
        &n,
        &_entries));
  EXPECT_ULEQ(0, n);
  delete(&_entries, n, string_delete);

  string_new_concat(&_filename, _dirname->_filename, "/file1.txt");
  file_create(_filename);
  EXPECT_TRUE(
      directory_try_read(
        _dirname->_filename,
        null,
        null,
        false,
        false,
        &n,
        &_entries));
  EXPECT_ULEQ(1, n);
  EXPECT_STREQ("file1.txt", _entries[0]);
  delete(&_entries, n, string_delete);

  file_remove(_filename);
  string_delete(&_filename);

  directory_remove(_dirname->_filename);
  temp_filename_delete(&_dirname);

  EXPECT_FALSE(
      directory_try_read(
        "/catdog/pig",
        null,
        null,
        false,
        false,
        &n,
        &_entries));
}

TEST(Directory, EnsureExists) {
  TempFilename *_dirname = null;

  temp_filename_new(&_dirname, "cat");
  EXPECT_FALSE(directory_exists(_dirname->_filename));
  directory_ensure_exists(_dirname->_filename);
  temp_filename_unlock(_dirname);
  EXPECT_TRUE(directory_exists(_dirname->_filename));
  directory_remove(_dirname->_filename);
  EXPECT_FALSE(directory_exists(_dirname->_filename));
  temp_filename_delete(&_dirname);
}

TEST(Directory, EnsureExistsForPath) {
  char *_filename = null;
  TempFilename *_dirname = null;

  temp_filename_new(&_dirname, "cat");
  string_new_concat(&_filename, _dirname->_filename, "/woof.txt");
  EXPECT_FALSE(directory_exists(_dirname->_filename));
  directory_ensure_exists_for_path(_filename);
  temp_filename_unlock(_dirname);
  EXPECT_TRUE(directory_exists(_dirname->_filename));
  directory_remove(_dirname->_filename);
  EXPECT_FALSE(directory_exists(_dirname->_filename));
  string_delete(&_filename);
  temp_filename_delete(&_dirname);
}

TEST(Directory, Remove) {
  TempFilename *_dirname = null;

  temp_filename_new(&_dirname, "cat");
  EXPECT_FALSE(directory_exists(_dirname->_filename));
  directory_ensure_exists(_dirname->_filename);
  temp_filename_unlock(_dirname);
  EXPECT_TRUE(directory_exists(_dirname->_filename));
  directory_remove(_dirname->_filename);
  EXPECT_FALSE(directory_exists(_dirname->_filename));
  temp_filename_delete(&_dirname);
}

TEST(Directory, Rename) {
  char *_filename = null;
  char *_filename_new = null;
  char *_filename_old = null;
  char *_dirname1 = null;
  char *_dirname1_new = null;
  char *_dirname2 = null;
  char *_dirname2_new = null;
  TempFilename *_dirname0 = null;
  TempFilename *_dirname0_new = null;

  temp_filename_new(&_dirname0, "cat");
  string_new_concat(&_dirname1, _dirname0->_filename, "/dog");
  string_new_concat(&_dirname2, _dirname1, "/woof");
  string_new_concat(&_filename, _dirname2, "/pig.txt");
  EXPECT_FALSE(directory_exists(_dirname0->_filename));
  EXPECT_FALSE(directory_exists(_dirname1));
  EXPECT_FALSE(directory_exists(_dirname2));
  EXPECT_FALSE(file_exists(_filename));
  directory_ensure_exists_for_path(_filename);
  temp_filename_unlock(_dirname0);
  EXPECT_TRUE(directory_exists(_dirname0->_filename));
  EXPECT_TRUE(directory_exists(_dirname1));
  EXPECT_TRUE(directory_exists(_dirname2));
  EXPECT_FALSE(file_exists(_filename));
  file_create(_filename);
  EXPECT_TRUE(file_exists(_filename));
  temp_filename_new(&_dirname0_new, "miaow");
  string_new_concat(&_dirname1_new, _dirname0_new->_filename, "/yelp");
  string_new_concat(&_dirname2_new, _dirname1_new, "/growl");
  string_new_concat(&_filename_new, _dirname2_new, "/snort.gif");
  string_new_concat(&_filename_old, _dirname2_new, "/pig.txt");
  EXPECT_FALSE(directory_exists(_dirname0_new->_filename));
  EXPECT_FALSE(directory_exists(_dirname1_new));
  EXPECT_FALSE(directory_exists(_dirname2_new));
  EXPECT_FALSE(file_exists(_filename_new));
  EXPECT_FALSE(file_exists(_filename_old));
  directory_rename(_dirname2, _dirname2_new);
  temp_filename_unlock(_dirname0_new);
  EXPECT_FALSE(directory_exists(_dirname0->_filename));
  EXPECT_FALSE(directory_exists(_dirname1));
  EXPECT_FALSE(directory_exists(_dirname2));
  EXPECT_FALSE(file_exists(_filename));
  EXPECT_TRUE(directory_exists(_dirname0_new->_filename));
  EXPECT_TRUE(directory_exists(_dirname1_new));
  EXPECT_TRUE(directory_exists(_dirname2_new));
  EXPECT_FALSE(file_exists(_filename_new));
  EXPECT_TRUE(file_exists(_filename_old));
  file_rename(_filename_old, _filename_new);
  EXPECT_TRUE(file_exists(_filename_new));
  EXPECT_FALSE(file_exists(_filename_old));
  file_remove(_filename_new);
  directory_remove(_dirname2_new);
  directory_remove(_dirname1_new);
  directory_remove(_dirname0_new->_filename);
  EXPECT_FALSE(directory_exists(_dirname0->_filename));
  EXPECT_FALSE(directory_exists(_dirname1));
  EXPECT_FALSE(directory_exists(_dirname2));
  EXPECT_FALSE(file_exists(_filename));
  EXPECT_FALSE(directory_exists(_dirname0_new->_filename));
  EXPECT_FALSE(directory_exists(_dirname1_new));
  EXPECT_FALSE(directory_exists(_dirname2_new));
  EXPECT_FALSE(file_exists(_filename_new));
  EXPECT_FALSE(file_exists(_filename_old));
  string_delete(&_filename_old);
  string_delete(&_filename_new);
  string_delete(&_dirname2_new);
  string_delete(&_dirname1_new);
  temp_filename_delete(&_dirname0_new);
  string_delete(&_filename);
  string_delete(&_dirname2);
  string_delete(&_dirname1);
  temp_filename_delete(&_dirname0);
}

TEST(Path, RenamePrune) {
  TempFilename *_dirname0 = null;
  TempFilename *_dirname0_new = null;
  char *_dirname1 = null;
  char *_dirname1_new = null;
  char *_dirname2 = null;
  char *_dirname2_new = null;
  char *_filename = null;
  char *_filename_new = null;

  temp_filename_new(&_dirname0, "cat");
  string_new_concat(&_dirname1, _dirname0->_filename, "/dog");
  string_new_concat(&_dirname2, _dirname1, "/woof");
  string_new_concat(&_filename, _dirname2, "/pig.txt");
  temp_filename_new(&_dirname0_new, "miaow");
  string_new_concat(&_dirname1_new, _dirname0_new->_filename, "/yelp");
  string_new_concat(&_dirname2_new, _dirname1_new, "/growl");
  string_new_concat(&_filename_new, _dirname2_new, "/snort.gif");

  directory_ensure_exists_for_path(_filename);
  temp_filename_unlock(_dirname0);
  temp_filename_unlock(_dirname0_new);
  file_create(_filename);
  EXPECT_TRUE(file_exists(_filename));
  path_rename(_filename, _filename_new);
  EXPECT_TRUE(file_exists(_filename_new));
  EXPECT_FALSE(file_exists(_filename));
  path_remove(_filename_new);
  EXPECT_FALSE(directory_exists(_dirname0->_filename));
  EXPECT_FALSE(directory_exists(_dirname1));
  EXPECT_FALSE(directory_exists(_dirname2));
  EXPECT_FALSE(file_exists(_filename));
  EXPECT_FALSE(directory_exists(_dirname0_new->_filename));
  EXPECT_FALSE(directory_exists(_dirname1_new));
  EXPECT_FALSE(directory_exists(_dirname2_new));
  EXPECT_FALSE(file_exists(_filename_new));

  directory_ensure_exists_for_path(_filename);
  file_create(_filename);
  path_prune_directory(_filename);
  EXPECT_TRUE(file_exists(_filename));
  file_remove(_filename);
  path_prune_directory(_filename);
  EXPECT_FALSE(file_exists(_filename));
  EXPECT_FALSE(directory_exists(_dirname0->_filename));
  EXPECT_FALSE(directory_exists(_dirname1));
  EXPECT_FALSE(directory_exists(_dirname2));

  string_delete(&_filename_new);
  string_delete(&_dirname2_new);
  string_delete(&_dirname1_new);
  temp_filename_delete(&_dirname0_new);
  string_delete(&_filename);
  string_delete(&_dirname2);
  string_delete(&_dirname1);
  temp_filename_delete(&_dirname0);
}

FINISH
